﻿//---------------------------------------------------------------------
// <copyright file="CollectionValueReaderTests.cs" company="Microsoft">
//      Copyright (C) Microsoft Corporation. All rights reserved. See License.txt in the project root for license information.
// </copyright>
//---------------------------------------------------------------------

namespace Microsoft.Test.Taupo.OData.Reader.Tests.Reader
{
    #region Namespaces
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using Microsoft.OData;
    using Microsoft.OData.Edm;
    using Microsoft.Test.OData.Utils.ODataLibTest;
    using Microsoft.Test.Taupo.Astoria.Common;
    using Microsoft.Test.Taupo.Astoria.Contracts.OData;
    using Microsoft.Test.Taupo.Common;
    using Microsoft.Test.Taupo.Contracts.EntityModel;
    using Microsoft.Test.Taupo.Contracts.EntityModel.Edm;
    using Microsoft.Test.Taupo.Contracts.Types;
    using Microsoft.Test.Taupo.Execution;
    using Microsoft.Test.Taupo.OData.Common;
    using Microsoft.Test.Taupo.OData.Common.Annotations;
    using Microsoft.Test.Taupo.OData.Contracts;
    using Microsoft.Test.Taupo.OData.Contracts.JsonLight;
    using Microsoft.Test.Taupo.OData.JsonLight;
    using Microsoft.Test.Taupo.OData.Reader.Tests;
    using Microsoft.VisualStudio.TestTools.UnitTesting;
    #endregion Namespaces

    /// <summary>
    /// Tests reading of various collection payloads.
    /// </summary>
    [TestClass, TestCase]
    public class CollectionValueReaderTests : ODataReaderTestCase
    {
        [InjectDependency]
        public IPayloadGenerator PayloadGenerator { get; set; }

        [InjectDependency]
        public PayloadReaderTestDescriptor.Settings Settings { get; set; }

        [TestMethod, TestCategory("Reader.Collections"), Variation(Description = "Verifies correct reading of collections with fully specified metadata.")]
        public void CollectionWithExpectedTypeAndMetadataTest()
        {
            IEnumerable<PayloadReaderTestDescriptor> testDescriptors = this.CreateCollectionPayloadsWithMetadata(true);
            testDescriptors = testDescriptors.Concat(this.CreateInvalidCollectionsWithTypeNames(true));

            // Wrap the value in a property and in different types of payloads
            testDescriptors = testDescriptors
                .Select((td, index) => td.InProperty("propertyName" + index))   // This adds the expected type annotation based on the type of the property value
                .SelectMany(td => this.PayloadGenerator.GenerateReaderPayloads(td));

            this.CombinatorialEngineProvider.RunCombinations(
                testDescriptors,
                this.ReaderTestConfigurationProvider.ExplicitFormatConfigurations,
                (testDescriptor, testConfiguration) =>
                {
                    var property = testDescriptor.PayloadElement as PropertyInstance;
                    testDescriptor.RunTest(testConfiguration);
                });
        }

        [TestMethod, TestCategory("Reader.Collections"), Variation(Description = "Verifies correct reading of collections without expected type but with metadata.")]
        public void CollectionWithoutExpectedTypeAndWithMetadataTest()
        {
            // For now only top-level property can do this.
            // TODO: Once we have open properties, these test cases apply to those as well, then probably move these to the top-level property
            // tests and share them from the open properties test, or possible keep both here.

            IEnumerable<PayloadReaderTestDescriptor> testDescriptors = this.CreateCollectionPayloadsWithMetadata(true);
            testDescriptors = testDescriptors.Concat(this.CreateInvalidCollectionsWithTypeNames(false));

            EdmModel model = new EdmModel();
            EdmComplexType itemComplexType = model.ComplexType("ItemComplexType").Property("stringProperty", EdmPrimitiveTypeKind.String);
            model = model.Fixup();

            testDescriptors = testDescriptors.Concat(new[]
                {
                    // No expected type specified, the one in the payload should be enough
                    new PayloadReaderTestDescriptor(this.Settings)
                    {
                        PayloadElement = PayloadBuilder.PrimitiveMultiValue(EntityModelUtils.GetCollectionTypeName("Edm.String")),
                        PayloadEdmModel = model,
                    },
                    new PayloadReaderTestDescriptor(this.Settings)
                    {
                        PayloadElement = PayloadBuilder.ComplexMultiValue(EntityModelUtils.GetCollectionTypeName("TestModel.ItemComplexType")),
                        PayloadEdmModel = model,
                    },

                    // Verify that the item type is inherited from the collection to its items if the item doesn't specify the type
                    new PayloadReaderTestDescriptor(this.Settings)
                    {
                        PayloadElement = PayloadBuilder.ComplexMultiValue(EntityModelUtils.GetCollectionTypeName("TestModel.ItemComplexType"))
                            .Item(PayloadBuilder
                                    .ComplexValue()
                                    .PrimitiveProperty("stringProperty", "test")
                                    .WithTypeAnnotation(itemComplexType)
                                    .AddAnnotation(new SerializationTypeNameTestAnnotation() { TypeName = null })),  // Add item which does not have the type name
                        PayloadEdmModel = model
                    },
                    new PayloadReaderTestDescriptor(this.Settings)
                    {
                        PayloadElement = PayloadBuilder.ComplexMultiValue(EntityModelUtils.GetCollectionTypeName("TestModel.ItemComplexType"))
                            .Item(PayloadBuilder
                                    .ComplexValue("TestModel.ItemComplexType")
                                    .PrimitiveProperty("stringProperty", "test")
                                    .WithTypeAnnotation(itemComplexType))  // Add an item which does have the type name
                            .Item(PayloadBuilder
                                    .ComplexValue()
                                    .PrimitiveProperty("stringProperty", "test")
                                    .WithTypeAnnotation(itemComplexType)
                                    .AddAnnotation(new SerializationTypeNameTestAnnotation() { TypeName = null })),  // Add item which does not have the type name
                        PayloadEdmModel = model
                    }
                });

            // Wrap the value in a top-level property without expected type (can't use the .InProperty here, since that would put the expected type on it)
            testDescriptors = testDescriptors.Select(td =>
                new PayloadReaderTestDescriptor(td)
                {
                    PayloadElement = PayloadBuilder.Property("propertyName", td.PayloadElement)
                });

            // Fill in type names for expected result from the type annotations
            testDescriptors = testDescriptors.Select(td =>
            {
                td.ExpectedResultNormalizers.Add(tc => FillTypeNamesFromTypeAnnotationsPayloadElementVisitor.Visit);
                return td;
            });

            this.CombinatorialEngineProvider.RunCombinations(
                testDescriptors,
                // The version dependent behavior tests are implemented in the format specific tests.
                this.ReaderTestConfigurationProvider.ExplicitFormatConfigurations,
                (testDescriptor, testConfiguration) =>
                {
                    testDescriptor.PayloadNormalizers.Add((tc) => tc.Format == ODataFormat.Json ? ReplaceExpectedTypeWithContextUriVisitor.VisitPayload : (Func<ODataPayloadElement, ODataPayloadElement>)null);

                    var property = testDescriptor.PayloadElement as PropertyInstance;
                    testDescriptor.RunTest(testConfiguration);
                });
        }

        [TestMethod, TestCategory("Reader.Collections"), Variation(Description = "Verifies correct reading of collections without typename, but with expected type and metadata.")]
        public void CollectionWithExpectedTypeAndMetadataButWithoutTypeNameTest()
        {
            IEnumerable<PayloadReaderTestDescriptor> testDescriptors = this.CreateCollectionPayloadsWithMetadata(false);

            // Wrap the value in a property and in different types of payloads
            testDescriptors = testDescriptors
                .Select((td, index) => td.InProperty("propertyName" + index))   // This adds the expected type annotation based on the type of the property value
                .SelectMany(td => this.PayloadGenerator.GenerateReaderPayloads(td));

            // Fill in type names for expected result from the type annotations
            testDescriptors = testDescriptors.Select(td =>
            {
                td.ExpectedResultNormalizers.Add(tc => FillTypeNamesFromTypeAnnotationsPayloadElementVisitor.Visit);
                return td;
            });

            this.CombinatorialEngineProvider.RunCombinations(
                testDescriptors,
                this.ReaderTestConfigurationProvider.ExplicitFormatConfigurations,
                (testDescriptor, testConfiguration) =>
                {
                    if (testConfiguration.Format == ODataFormat.Json && testConfiguration.MessageReaderSettings.BaseUri == null)
                    {
                        testConfiguration.MessageReaderSettings.BaseUri = new Uri("http://odata.org/");
                    }
                    var property = testDescriptor.PayloadElement as PropertyInstance;
                    testDescriptor.RunTest(testConfiguration);
                });
        }

        [TestMethod, TestCategory("Reader.Collections"), Variation(Description = "Verifies correct error handling of collections containing heterogenous items")]
        public void CollectionWithHeterogenousItemsErrorTest()
        {
            EdmModel model = new EdmModel();

            var complexType1 = model.ComplexType("ComplexTypeWithStringAndInteger32")
                .Property("Property1", EdmCoreModel.Instance.GetString(true) as EdmTypeReference)
                .Property("Property2", EdmCoreModel.Instance.GetInt32(false) as EdmTypeReference);

            var complexType2 = model.ComplexType("ComplexTypeWithStringAndDateTime")
                .Property("Property1", EdmCoreModel.Instance.GetString(true) as EdmTypeReference)
                .Property("Property2", EdmPrimitiveTypeKind.DateTimeOffset);

            model.Fixup();

            IEnumerable<PayloadReaderTestDescriptor> testDescriptors = new[]
            {
                // Primitive collection containing items of different primitive types
                new PayloadReaderTestDescriptor(this.Settings)
                {
                    PayloadElement = PayloadBuilder.PrimitiveMultiValue(EntityModelUtils.GetCollectionTypeName("Edm.Int32")).Item(1).Item(true).Item(2).WithTypeAnnotation(EdmCoreModel.GetCollection(EdmCoreModel.Instance.GetInt32(false))),
                    PayloadEdmModel = model,
                    ExpectedResultCallback = tc => new PayloadReaderTestExpectedResult(this.Settings.ExpectedResultSettings)
                    {
                        ExpectedException = ODataExpectedExceptions.ODataException("ReaderValidationUtils_CannotConvertPrimitiveValue", "True", "Edm.Int32")
                    }
                },
                // Complex collection containing items of different complex type (correct type attribute value)
                new PayloadReaderTestDescriptor(this.Settings)
                {
                    PayloadElement = PayloadBuilder.ComplexMultiValue(EntityModelUtils.GetCollectionTypeName(complexType1.FullName()))
                        .Item(PayloadBuilder.ComplexValue(complexType1.FullName()).PrimitiveProperty("Property1", "Foo").PrimitiveProperty("Property2", -1))
                        .Item(PayloadBuilder.ComplexValue(complexType2.FullName()).PrimitiveProperty("Property1", "Foo").PrimitiveProperty("Property2", DateTimeOffset.Now))
                        .WithTypeAnnotation(EdmCoreModel.GetCollection(complexType1.ToTypeReference())),
                    PayloadEdmModel = model,
                    ExpectedException = ODataExpectedExceptions.ODataException("ValidationUtils_IncompatibleType", "TestModel.ComplexTypeWithStringAndDateTime", "TestModel.ComplexTypeWithStringAndInteger32"),
                },
                // Complex collection containing items of different complex type (incorrect type attribute value)
                new PayloadReaderTestDescriptor(this.Settings)
                {
                    PayloadElement = PayloadBuilder.ComplexMultiValue(EntityModelUtils.GetCollectionTypeName(complexType2.FullName()))
                        .Item(PayloadBuilder.ComplexValue(complexType2.FullName()).PrimitiveProperty("Property1", "Foo").PrimitiveProperty("Property2", -1))
                        .Item(PayloadBuilder.ComplexValue(complexType2.FullName()).PrimitiveProperty("Property1", "Foo").PrimitiveProperty("Property2", DateTimeOffset.Now))
                        .WithTypeAnnotation(EdmCoreModel.GetCollection(complexType2.ToTypeReference())),
                    PayloadEdmModel = model,
                    ExpectedResultCallback = tc => new PayloadReaderTestExpectedResult(this.Settings.ExpectedResultSettings)
                    {
                        ExpectedException = ODataExpectedExceptions.ODataException("ReaderValidationUtils_CannotConvertPrimitiveValue", "-1", "Edm.DateTimeOffset")
                    }
                },
            };

            this.CombinatorialEngineProvider.RunCombinations(
                 testDescriptors,
                 this.ReaderTestConfigurationProvider.ExplicitFormatConfigurations,
                 (testDescriptor, testConfiguration) =>
                 {
                     testDescriptor = testDescriptor.InProperty("RootProperty");
                     testDescriptor.RunTest(testConfiguration);
                 });
        }

        private IEnumerable<PayloadReaderTestDescriptor> CreateCollectionPayloadsWithMetadata(bool withTypeNames)
        {
            // Start with the standard set of collections
            IEnumerable<PayloadReaderTestDescriptor> testDescriptors = PayloadReaderTestDescriptorGenerator.CreateCollectionTestDescriptors(this.Settings, withTypeNames);

            // Add collections with all of the primitive values (except null)
            testDescriptors = testDescriptors.Concat(PayloadReaderTestDescriptorGenerator.CreatePrimitiveValueTestDescriptors(this.Settings, false)
                .Where(primitivePayload => ((PrimitiveValue)primitivePayload.PayloadElement).ClrValue != null)
                .Select(primitivePayload =>
                {
                    IEdmTypeReference edmType = primitivePayload.PayloadElement.GetAnnotation<EntityModelTypeAnnotation>().EdmModelType;
                    var primitiveCollection = PayloadBuilder.PrimitiveMultiValue(withTypeNames ? EntityModelUtils.GetCollectionTypeName(edmType.FullName()) : null)
                            .Item(((PrimitiveValue)primitivePayload.PayloadElement).ClrValue)
                            .WithTypeAnnotation(EdmCoreModel.GetCollection(edmType));
                    if (!withTypeNames)
                    {
                        primitiveCollection.AddAnnotation(new SerializationTypeNameTestAnnotation() { TypeName = null });
                    }
                    return new PayloadReaderTestDescriptor(primitivePayload)
                    {
                        PayloadElement = primitiveCollection,
                    };
                }));

            // Add collections with all of the complex values (except null)
            testDescriptors = testDescriptors.Concat(PayloadReaderTestDescriptorGenerator.CreateComplexValueTestDescriptors(this.Settings, withTypeNames, false)
                .Where(complexPayload => !((ComplexInstance)complexPayload.PayloadElement).IsNull)
                .Select(complexPayload =>
                {
                    IEdmTypeReference edmComplexType = complexPayload.PayloadElement.GetAnnotation<EntityModelTypeAnnotation>().EdmModelType;
                    var complexCollection = PayloadBuilder.ComplexMultiValue(withTypeNames ? EntityModelUtils.GetCollectionTypeName(edmComplexType.FullName()) : null)
                            .Item((ComplexInstance)complexPayload.PayloadElement)
                            .WithTypeAnnotation(EdmCoreModel.GetCollection(edmComplexType));
                    if (!withTypeNames)
                    {
                        complexCollection.AddAnnotation(new SerializationTypeNameTestAnnotation() { TypeName = null });
                    }
                    return new PayloadReaderTestDescriptor(complexPayload)
                    {
                        PayloadElement = complexCollection,
                    };
                }));

            return testDescriptors;
        }

        public IEnumerable<PayloadReaderTestDescriptor> CreateInvalidCollectionsWithTypeNames(bool expectedTypeWillBeUsed)
        {
            EdmModel model = new EdmModel();
            EdmComplexType itemComplexType = model.ComplexType("ItemComplexType");
            model.ComplexType("ExtraComplexType");
            model = model.Fixup();

            // Add invalid cases
            var testDescriptors = new[]
            {
                // Invalid collection type name
                new PayloadReaderTestDescriptor(this.Settings)
                {
                    PayloadElement = PayloadBuilder.PrimitiveMultiValue("")
                        .WithTypeAnnotation(EdmCoreModel.GetCollection(EdmCoreModel.Instance.GetInt32(false))),
                    PayloadEdmModel = model,
                    ExpectedException = ODataExpectedExceptions.ODataException("ValidationUtils_UnrecognizedTypeName", string.Empty),
                },
                new PayloadReaderTestDescriptor(this.Settings)
                {
                    PayloadElement = PayloadBuilder.PrimitiveMultiValue(EntityModelUtils.GetCollectionTypeName(""))
                        .WithTypeAnnotation(EdmCoreModel.GetCollection(EdmCoreModel.Instance.GetInt32(false))),
                    PayloadEdmModel = model,
                    ExpectedException = ODataExpectedExceptions.ODataException("ValidationUtils_UnrecognizedTypeName", EntityModelUtils.GetCollectionTypeName("")),
                },
                // Invalid collection type name
                new PayloadReaderTestDescriptor(this.Settings)
                {
                    PayloadElement = PayloadBuilder.PrimitiveMultiValue("collection(Edm.Int32)")
                        .WithTypeAnnotation(EdmCoreModel.GetCollection(EdmCoreModel.Instance.GetInt32(false))),
                    PayloadEdmModel = model,
                    ExpectedException = ODataExpectedExceptions.ODataException("ValidationUtils_UnrecognizedTypeName", "collection(Edm.Int32)"),
                },
                new PayloadReaderTestDescriptor(this.Settings)
                {
                    PayloadElement = PayloadBuilder.PrimitiveMultiValue("foo")
                        .WithTypeAnnotation(EdmCoreModel.GetCollection(EdmCoreModel.Instance.GetInt32(false))),
                    PayloadEdmModel = model,
                    ExpectedException = ODataExpectedExceptions.ODataException("ValidationUtils_UnrecognizedTypeName", "foo"),
                },

                // Non existant type name
                new PayloadReaderTestDescriptor(this.Settings)
                {
                    PayloadElement = PayloadBuilder.PrimitiveMultiValue(EntityModelUtils.GetCollectionTypeName("TestModel.NonExistant"))
                        .WithTypeAnnotation(EdmCoreModel.GetCollection(EdmCoreModel.Instance.GetInt32(false))),
                    PayloadEdmModel = model,
                    ExpectedException = ODataExpectedExceptions.ODataException("ValidationUtils_UnrecognizedTypeName", EntityModelUtils.GetCollectionTypeName("TestModel.NonExistant")),
                },

                // Type of the item differs from the type of the collection
                new PayloadReaderTestDescriptor(this.Settings)
                {
                    PayloadElement = PayloadBuilder.PrimitiveMultiValue(EntityModelUtils.GetCollectionTypeName("Edm.String"))
                        .Item(-42)
                        .WithTypeAnnotation(EdmCoreModel.GetCollection(EdmCoreModel.Instance.GetString(true))),
                    PayloadEdmModel = model,
                    ExpectedResultCallback = tc =>
                        new PayloadReaderTestExpectedResult(this.Settings.ExpectedResultSettings)
                        {
                            ExpectedException = ODataExpectedExceptions.ODataException("ReaderValidationUtils_CannotConvertPrimitiveValue", "-42", "Edm.String")
                        },
                },
                new PayloadReaderTestDescriptor(this.Settings)
                {
                    PayloadElement = PayloadBuilder.ComplexMultiValue(EntityModelUtils.GetCollectionTypeName("TestModel.ItemComplexType"))
                        .Item(PayloadBuilder.ComplexValue("TestModel.ExtraComplexType"))
                        .WithTypeAnnotation(EdmCoreModel.GetCollection(itemComplexType.ToTypeReference())),
                    PayloadEdmModel = model,
                    ExpectedException = ODataExpectedExceptions.ODataException("ValidationUtils_IncompatibleType", "TestModel.ExtraComplexType", "TestModel.ItemComplexType"),
                }
            };

            if (expectedTypeWillBeUsed)
            {
                testDescriptors = testDescriptors.Concat(new[]
                    {
                        // Type differs from the declared/expected type
                        new PayloadReaderTestDescriptor(this.Settings)
                        {
                            PayloadElement = PayloadBuilder.PrimitiveMultiValue(EntityModelUtils.GetCollectionTypeName("Edm.String"))
                                .WithTypeAnnotation(EdmCoreModel.GetCollection(EdmCoreModel.Instance.GetInt32(false))),
                            PayloadEdmModel = model,
                            ExpectedException = ODataExpectedExceptions.ODataException("ValidationUtils_IncompatibleType", EntityModelUtils.GetCollectionTypeName("Edm.String"), EntityModelUtils.GetCollectionTypeName("Edm.Int32")),
                        },
                    }).ToArray();
            }

            foreach (var testDescriptor in testDescriptors)
            {
                testDescriptor.PayloadNormalizers.Add((tc) => tc.Format == ODataFormat.Json ? AddJsonLightTypeAnnotationToCollectionsVisitor.Normalize : (Func<ODataPayloadElement, ODataPayloadElement>)null);
            }

            return testDescriptors;

        }
    }

    #region ReplaceExpectedTypeWithContextUriVisitor
    /// <summary>
    /// Visitor for generating json light context uri annotations to replace any expected type annotations (that
    /// may have been generated by the Json Light fixup)
    /// </summary>
    internal class ReplaceExpectedTypeWithContextUriVisitor : ODataPayloadElementVisitorBase
    {
        private readonly Stack<ODataPayloadElement> payloadElementStack = new Stack<ODataPayloadElement>();

        public static ODataPayloadElement VisitPayload(ODataPayloadElement payloadElement)
        {
            new ReplaceExpectedTypeWithContextUriVisitor().Recurse(payloadElement);
            return payloadElement;
        }

        public override void Visit(ComplexProperty payloadElement)
        {
            base.Visit(payloadElement);
            this.ReplaceExpectedTypeAnnotationIfRootElement(payloadElement);
        }

        public override void Visit(PrimitiveMultiValueProperty payloadElement)
        {
            base.Visit(payloadElement);
            this.ReplaceExpectedTypeAnnotationIfRootElement(payloadElement);
        }

        public override void Visit(ComplexMultiValueProperty payloadElement)
        {
            base.Visit(payloadElement);
            this.ReplaceExpectedTypeAnnotationIfRootElement(payloadElement);
        }

        public override void Visit(PrimitiveProperty payloadElement)
        {
            base.Visit(payloadElement);
            this.ReplaceExpectedTypeAnnotationIfRootElement(payloadElement);
        }

        protected override void Recurse(ODataPayloadElement element)
        {
            try
            {
                this.payloadElementStack.Push(element);
                base.Recurse(element);
            }
            finally
            {
                this.payloadElementStack.Pop();
            }
        }

        private void ReplaceExpectedTypeAnnotationIfRootElement(ODataPayloadElement payloadElement)
        {
            if (this.payloadElementStack.Count == 2)
            {
                var expectedTypeAnnotation = payloadElement.GetAnnotation<ExpectedTypeODataPayloadElementAnnotation>();
                if (expectedTypeAnnotation != null)
                {
                    payloadElement.WithContextUri(JsonLightODataPayloadElementExtensions.BuildContextUri(payloadElement.ElementType, JsonLightConstants.DefaultMetadataDocumentUri.OriginalString, expectedTypeAnnotation));
                    payloadElement.RemoveAnnotations(typeof(ExpectedTypeODataPayloadElementAnnotation));
                }
            }
        }
    }
    #endregion

    #region AddJsonLightTypeAnnotationToCollectionsVisitor
    /// <summary>
    /// Visitor for adding a Json Light type property annotation to collection values.
    /// </summary>
    public class AddJsonLightTypeAnnotationToCollectionsVisitor : ODataPayloadElementVisitorBase
    {
        private readonly Stack<ODataPayloadElement> payloadElementStack = new Stack<ODataPayloadElement>();

        public static ODataPayloadElement Normalize(ODataPayloadElement payloadElement)
        {
            new AddJsonLightTypeAnnotationToCollectionsVisitor().Recurse(payloadElement);
            return payloadElement;
        }

        public override void Visit(ComplexMultiValue payloadElement)
        {
            base.Visit(payloadElement);
            this.AddTypeAnnotation(payloadElement, payloadElement.FullTypeName);
        }

        public override void Visit(PrimitiveMultiValue payloadElement)
        {
            base.Visit(payloadElement);
            this.AddTypeAnnotation(payloadElement, payloadElement.FullTypeName);
        }

        protected override void Recurse(ODataPayloadElement element)
        {
            try
            {
                this.payloadElementStack.Push(element);
                base.Recurse(element);
            }
            finally
            {
                this.payloadElementStack.Pop();
            }
        }

        private void AddTypeAnnotation(ODataPayloadElement payloadElement, string typeName)
        {
            var propertyInstance = this.payloadElementStack.ElementAt(1) as PropertyInstance;
            if (propertyInstance != null)
            {
                propertyInstance.WithPropertyAnnotation(JsonLightConstants.ODataTypeAnnotationName, typeName);
            }
        }
    }
    #endregion
}
